"""
Tests automation sequences
"""
import logging
import time
import unittest
import random

from hein_control import ConfiguredAction, EvalAction, ConfiguredDecision, Automation
from hein_control.sequencing import ConfiguredAutomation


def wait_long():
    time.sleep(0.05)
    logging.info('done waiting')


def greater_than_5(number) -> bool:
    if number > 5:
        return True
    else:
        return False


class VariableTracker:
    def __init__(self):
        self.hello_world = ''


class TestAutomationSequencing(unittest.TestCase):

    def test_basic_automation(self):  # executing in real time
        """tests basic automation"""
        action_1 = ConfiguredAction(wait_long)
        action_2 = ConfiguredAction(wait_long)
        action_1.next_action = action_2
        config_auto = ConfiguredAutomation(first_action=action_1)
        automation = config_auto.get_tracked_from_config()
        automation()
        self.assertTrue(automation.in_progress)
        self.assertFalse(automation.complete)
        time.sleep(0.05)
        self.assertTrue(automation.in_progress)
        self.assertFalse(automation.complete)
        time.sleep(0.07)
        self.assertFalse(automation.in_progress)
        self.assertTrue(automation.complete)
        for ta in automation.executed_actions:
            self.assertTrue(ta.started)
            self.assertFalse(ta.error)
            self.assertEqual(ta.status_code, 3)


class TestTurbidityEx(unittest.TestCase):
    """tests automation with a turbidity example"""

    # functions for a workflow for a turbidity experiment using log statements
    @staticmethod
    @ConfiguredAction
    def set_up_camera_config(**kwargs):
        time.sleep(0.01)
        logging.info('set up camera, start experiment')

    @staticmethod
    @ConfiguredAction
    def measure_turbidity(n_photos: int, **kwargs):  # pass in the number of photos to take
        def take_n_photos(n_photos: int):
            logging.info(f'take {n_photos} photos')
            return [i for i in range(n_photos)]

        def analyze_image(img,):
            logging.info(f'analyzing image {img}')

        imgs = take_n_photos(n_photos=n_photos)
        logging.info('analyzing a batch of images')
        for i in imgs:
            analyze_image(i)

    @staticmethod
    def is_dissolved(**kwargs) -> bool:
        choice = random.choice([True, False])
        return choice

    @staticmethod
    def is_turbid(**kwargs) -> bool:
        # for the decision action, there needs to be the case where if is_dissolved returns False,
        # there needs to be another logic check (this one) that will return true
        return True

    @staticmethod
    @ConfiguredAction
    def alert_dissolved_config(**kwargs):
        logging.info('yay, everything has dissolved')

    @staticmethod
    @ConfiguredAction
    def end_experiment_config(**kwargs):
        logging.info('end experiment')

    def test_turbidity_automation(self):  # creating and running a turbidity experiment
        """tests an automation example based on turbidity analysis"""
        # create decision action to check if an experiment should continue based on turbidity or not
        is_dissolved = ConfiguredAction(EvalAction(self.is_dissolved), next_action=self.alert_dissolved_config)
        is_turbid = ConfiguredAction(EvalAction(self.is_turbid), next_action=self.measure_turbidity)
        check_turbidity_config = ConfiguredDecision(is_dissolved, is_turbid)

        # set up next actions for the configured action functions
        self.set_up_camera_config.next_action = self.measure_turbidity  # set up camera the measure turbidity
        self.measure_turbidity.next_action = check_turbidity_config  # after measuring turbidity, check if things
        # have dissolved or not
        self.alert_dissolved_config.next_action = self.end_experiment_config  # after alerting the user, end the
        # experiment

        n_photos = 2
        turbidity_exp = ConfiguredAutomation(first_action=self.set_up_camera_config)
        self.assertEqual(self.set_up_camera_config,
                         turbidity_exp.first_action)
        # first replicate, passing in kwargs when calling the automation
        logging.info('first replicate')
        replicate_1 = turbidity_exp.get_tracked_from_config()
        self.assertEqual(None,
                         replicate_1.executed_actions)
        self.assertFalse(replicate_1.triggered)
        self.assertEqual({},
                         replicate_1.kwargs)
        replicate_1(**{'n_photos': n_photos})
        replicate_1.wait_for_completion()
        self.assertTrue(replicate_1.triggered)
        tracked_action_list_1 = replicate_1.executed_actions
        logging.info(tracked_action_list_1)
        self.assertNotEqual(None,
                            tracked_action_list_1)
        self.assertNotEqual(0,
                            len(tracked_action_list_1))

        # second replicate, passing in kwargs when creating the tracked config
        n_photos = 10
        logging.info('second replicate')
        replicate_2 = turbidity_exp.get_tracked_from_config(**{'n_photos': n_photos})
        self.assertEqual(None,
                         replicate_2.executed_actions)
        self.assertFalse(replicate_2.triggered)
        self.assertEqual({'n_photos': 10},
                         replicate_2.kwargs)
        replicate_2()
        replicate_2.wait_for_completion()
        tracked_action_list_2 = replicate_2.executed_actions
        logging.info(tracked_action_list_2)
        self.assertTrue(replicate_2.triggered)
        self.assertNotEqual(None,
                            tracked_action_list_2)
        self.assertNotEqual(0,
                            len(replicate_1.executed_actions))





"""mixin classes for use across the package"""

from .status import Status
from .config import Configuration
from .reg import InstanceRegistry, ClassProperty

__all__ = [
    'Status',
    'Configuration',
    'InstanceRegistry',
    'ClassProperty',
]



import datetime
import logging
import threading
import time
import warnings
from functools import wraps
from typing import List, Union, Callable

from .basic import Action, ActionList
from ..mixin import Status, Configuration, InstanceRegistry


logger = logging.getLogger(__name__)


class ConfiguredAction(InstanceRegistry, Configuration, Status):
    UUID_PREFIX = 'ca-'

    def __init__(self,
                 action: Union[Callable, Action, str],
                 *args,
                 configuration_name: str = None,
                 next_action: Union[str, 'ConfiguredAction'] = None,
                 **kwargs,
                 ):
        """
        An action that is configured for execution. Arguments and Keyword arguments may be specified and will
        be handed to the action on execution unless overridden by the user. (Using kwargs over args is strongly
        recommended.)

        :param action: action to perform (callable)
        :param args: arguments for the action
        :param configuration_name: optional name for the configuration (for key retrieval)
        :param next_action: optional next action for sequence building
        :param kwargs: keyword arguments for the action
        """
        # todo ability to retrieve the executions of a given configured action
        InstanceRegistry.__init__(self)
        self._action: Action = None
        self._config_name: str = None
        self._next_action = None
        # register action and store attribute
        self.action = action
        Configuration.__init__(
            self,
            *self.action.default_args,
            **self.action.default_kwargs,
        )
        Status.__init__(self)
        self.update_args(*args)
        self.update_kwargs(**kwargs)
        # generate name if not specified
        if configuration_name is None:
            configuration_name = self._generate_unique_name(self.action)
        else:
            if configuration_name in self.get_configuration_names():
                raise NameError(f'The configuration name "{configuration_name}" is already used, please choose a '
                                f'unique name for the configuration. ')
        self.name = configuration_name
        if next_action is not None:
            self.next_action = next_action

    def __str__(self):
        return f'{self.__class__.__name__} {self.name}'

    def __repr__(self):
        return self.__str__()

    def __call__(self, *args, **kwargs):
        return self.action(*args, **kwargs)

    def __copy__(self):
        return self.__class__(
            self.action,
            self.next_action,
            self.args,
            self.name,
            self.kwargs,
        )

    def __deepcopy__(self, memodict={}):
        return self.__copy__()

    @property
    def action(self) -> Action:
        """The action to be tracked"""
        return self._action

    @action.setter
    def action(self, value: Union[Callable, Action, str]):
        if value is not None:
            self._action = Action.register_action(value)

    @property
    def name(self) -> str:
        """name for the configuration"""
        return self._config_name

    @name.setter
    def name(self, value: str):
        if value == self.name:
            return
        if value is not None:
            # check if already defined
            if value in self.get_configuration_names():
                raise NameError(f'The configuration name "{value}" is already used, please choose a '
                                f'unique name for the configuration. ')
            self._config_name = value

    @property
    def next_action(self) -> 'ConfiguredAction':
        """The next action in the sequence (for contextual action linking)"""
        return self._next_action

    @next_action.setter
    def next_action(self,
                    value: Union[str, 'ConfiguredAction'],
                    ):
        if value is None:
            self._next_action = None
            return
        # retrieve instance by key if not defined
        if isinstance(value, ConfiguredAction) is False:
            value = self.class_instance_by_id(value)
        self._next_action = value

    @next_action.deleter
    def next_action(self):
        self._next_action = None

    @property
    def previous_actions(self) -> List['ConfiguredAction']:
        """returns a list of configured actions which have this instance as a next step"""
        return [
            config
            for config in self.class_registry
            if config.next_action is self
        ]

    @classmethod
    def _generate_unique_name(cls, action: Action) -> str:
        """
        Generates a unique numbered name based on an Action instance. The resulting name will be unique in the
        registry.

        :param action: action to base naming off of
        :return: generated name
        """
        i = 2
        current_names = cls.get_configuration_names()
        if action.name not in current_names:
            return action.name
        while True:
            target_name = f'{action.name}_{i}'
            if f'{action.name}_{i}' not in current_names:
                return target_name
            i += 1

    @staticmethod
    def _match_instance_key(inst: "ConfiguredAction", identifier: str) -> bool:
        """
        Checks the provided identifier key against the instance provided, returning True if matching and False if not

        :param inst: InstanceRegistry instance
        :param identifier: string identifier
        :return: match
        """
        if inst._uuid == identifier:
            return True
        elif inst.uuid == identifier:
            return True
        elif inst.name == identifier:  # match configuration name
            return True
        return False

    @classmethod
    def get_configuration_by_name(cls, configuration_name: str) -> 'ConfiguredAction':
        """
        Retrieves a ConfiguredAction by name

        :param configuration_name: name identifier of configuration
        :return: ConfiguredAction
        """
        warnings.warn(
            'the get_configuration_by_name method has been deprecated, use class_instance_by_id',
            DeprecationWarning,
            stacklevel=2,
        )
        return cls.class_instance_by_id(configuration_name)

    @classmethod
    def get_configuration_names(cls) -> List[str]:
        """
        Retrieves the names of all configurations.

        Warning: return will be sorted, but order does not correspond to order in registry (indexing based on this
        return is not recommended).
        """
        return sorted([config.name for config in cls.class_registry if config.name is not None])

    @classmethod
    def get_configurations_of_action(cls, action: Union[str, Action]) -> List['ConfiguredAction']:
        """
        Retrieves all of the Configurations associated with the specified action

        :param action: action to reference
        :return: list of ConfiguredAction instances
        """
        if isinstance(action, Action) is False:
            action = Action.register_action(action)
        return [
            config
            for config in cls.class_registry
            if config.action is action
        ]

    @classmethod
    def remove_configuration_from_registry(cls, configuration: Union[str, 'ConfiguredAction']):
        """
        Removes the specified configuration from the configuration registry

        :param configuration: name or instance of the configuration
        """
        raise NotImplementedError('this method is not currently implemented')

    @classmethod
    def ensure_configuration(cls,
                             configuration: Union[str, Callable, Action, 'ConfiguredAction'],
                             ) -> 'ConfiguredAction':
        """
        Ensures that the provided configuration is a class instance, instantiating or converting as necessary.

        :param configuration: configuration instance, name, or key
        :raises ValueError: if a string is provided and an action is not registered
        :return: associated ConfiguredAction
        """
        # if already a configuration, return
        if isinstance(configuration, ConfiguredAction):
            return configuration
        # if a string key, retrieve
        elif type(configuration) is str:
            try:
                # try to retrieve configuration by name
                return ConfiguredAction.class_instance_by_id(configuration)
            except NameError:
                # attempt to create a configuration from a registered action
                return ConfiguredAction(
                    Action.register_action(configuration)
                )
        # if an action, create
        elif isinstance(configuration, Action):
            # if an action was provided, create configuration
            return ConfiguredAction(configuration)
        # if a callable, create an action
        elif callable(configuration):
            try:  # attempt to return an existing configuration of the function
                return ConfiguredAction.class_instance_by_id(configuration.__name__)
            except NameError:  # if undefined, register action and create configuration
                # if a callable was provided, register
                return ConfiguredAction(Action.register_action(configuration))
        raise ValueError(f'the value {configuration} could not be interpreted as a configuration')

    def duplicate_configuration(self, new_name: str = None) -> 'ConfiguredAction':
        """
        Duplicates a configuration with the provided name.

        :param new_name: new name for the generated ConfiguredAction instance
        :return: duplicated configuration
        """
        if new_name is None:
            new_name = self._generate_unique_name(self.action)
        return self.__class__(
            action=self.action,
            configuration_name=new_name,
            next_action=self.next_action,
            **self.kwargs,
        )

    def get_tracked_from_config(self, **kwargs) -> "TrackedAction":
        """
        Generates a tracked instance from the configured action

        :param kwargs: additional kwargs that differ from the instance configuration
        :return: TrackedAction (or subclass) instance.
        """
        return TrackedAction(
            configuration=self,
            **kwargs,
        )

    def as_dict(self) -> dict:
        """dictionary of relevant information"""
        if self.next_action is not None:
            next_action_name = self.next_action.name
        else:
            next_action_name = None
        out = {
            'name': self.name,
            'arguments': self.args,
            'keyword_arguments': self.kwargs,
            'status_code': self.status_code,
            'status': self.status_string,
            'next_action': next_action_name,
        }
        return out


class ConfiguredActionList(ActionList):
    def __init__(self, *actions: Union[Action, ConfiguredAction]):
        """
        A manager for a list of Configured Actions.

        :param actions: actions to sequence
        """
        self._action_list: List[ConfiguredAction] = []
        super().__init__(*actions)

    @staticmethod
    def _ensure_type(action: Union[str, Action, ConfiguredAction]) -> ConfiguredAction:
        """Ensures that the provided action is of the correct type for the list"""
        if isinstance(action, ConfiguredAction):
            return action
        elif type(action) is str:
            return ConfiguredAction.class_instance_by_id(action)
        else:
            return ConfiguredAction(action)


def ensure_triggered(fn: Callable):
    """decorator which will do the thing you want"""
    @wraps(fn)
    def decorated(self: 'TrackedAction', *args, **kwargs):
        if self.triggered is False:
            raise AttributeError(
                f'{fn.__name__} may not be accessed until after the {self.__class__.__name__} instance '
                f'is triggered or called'
            )
        return fn(self, *args, **kwargs)
    return decorated


class TrackedAction(InstanceRegistry, Configuration, Status):
    UUID_PREFIX = 'ta-'

    def __init__(self,
                 configuration: Union[str, Callable, Action, ConfiguredAction],
                 *args,
                 **kwargs,
                 ):
        """
        Creates a wrapped action which will track the time of start, completion, and return of the action.

        :param configuration: configuration to reference or action to perform (callable)
        :param args: arguments for the instance's action execution
        :param kwargs: keyword arguments for the instance's action execution
        """
        if isinstance(configuration, TrackedAction):
            raise ValueError('a TrackedAction instance may not be created from another TrackedAction instance, '
                             'use .copy_configuration instance method')
        InstanceRegistry.__init__(self)
        Configuration.__init__(self)
        Status.__init__(self)
        self._triggered = False
        self._configuration: ConfiguredAction = None
        self.configuration = configuration
        self.update_args(*args)
        self.update_kwargs(**kwargs)

        # error state tracker
        self.error: bool = False
        self.error_details = None
        self._time_started = None
        self._time_completed = None
        self._action_return = None

    def __call__(self, *args, **kwargs):
        if self._triggered is True:
            raise RuntimeError(f'multiple executions of a {self.__class__.__name__} instance is not permitted')
        self._triggered = True
        # update args and kwargs and store
        self.update_args(*args)
        self.update_kwargs(**kwargs)
        logger.debug('beginning action execution')
        self.freeze = True   # freeze settings
        self._time_started = time.time()  # set start time
        try:
            self._action_return = self.configuration.action(
                *self.args,
                **self.kwargs
            )
            logger.debug(f'action {self.configuration.action.name} completed successfully')
        except Exception as e:
            logger.error(f'error encountered when executing action {self.configuration.name}: {e}')
            self.error = True
            self.error_details = e
        self._time_completed = time.time()  # save completed time

    def __repr__(self):
        # todo have a separate str which indicates pending/triggered/complete as in timepoint
        strings = []
        args = self._arg_string
        if args != '':
            strings.append(args)
        kw = self._kwarg_string
        if kw != '':
            strings.append(kw)
        return f'{self.__class__.__name__}: {self.configuration.action.name}({", ".join(strings)})'

    def trigger(self, *args, **kwargs):
        """trigger execution of the tracked action"""
        self.__call__(*args, **kwargs)

    @property
    def triggered(self) -> bool:   # todo decide which of these to keep
        """whether the time point has been triggered"""
        return self._triggered

    @property
    def started(self) -> bool:
        """whether the timepoint has been started"""
        return self._triggered

    @property
    def action(self) -> Action:
        """configured action associated with the TrackedAction instances"""
        warnings.warn(  # v 6.0.0
            'the action attribute has been deprecated, access .configuration.action instead',
            DeprecationWarning,
            stacklevel=2,
        )
        return self.configuration.action

    @property
    def configuration(self) -> ConfiguredAction:
        """the configuration associated with the tracked action"""
        return self._configuration

    def _ensure_unfrozen(fn: Callable):
        """decorator which will do the thing you want"""
        @wraps(fn)
        def decorated(self: 'TrackedAction', *args, **kwargs):
            if self._freeze is True:
                raise ValueError(f'the instance is frozen and setting or updating of the args or kwargs is disabled')
            return fn(self, *args, **kwargs)
        return decorated

    @configuration.setter
    @_ensure_unfrozen
    def configuration(self, value: Union[str, Callable, Action, ConfiguredAction]):
        value = ConfiguredAction.ensure_configuration(value)
        self._configuration = value
        self.update_args(*self._configuration.args)
        self.update_kwargs(**self._configuration.kwargs)

    @property
    def time_started(self) -> float:
        """time stamp when the action was started"""
        return self._time_started

    @property
    def time_completed(self) -> float:
        """time stamp when the action was completed"""
        return self._time_completed

    @property
    def started_timestamp(self) -> datetime.datetime:
        """timestamp for when the action was started"""
        if self.time_started is not None:
            return datetime.datetime.fromtimestamp(self.time_started)

    @property
    def action_duration(self) -> float:
        """task duration (s)"""
        if self._time_completed is not None:
            return self._time_completed - self._time_started

    @property
    def method_return(self):
        """DEPRECATED retrieve action return"""
        warnings.warn(  # v 5.5.0
            'method_return is deprecated, use action_return instead',
            DeprecationWarning,
            stacklevel=2,
        )
        return self._action_return

    @property
    @ensure_triggered
    def action_return(self):
        """the return of the method once complete"""
        return self._action_return

    @property
    def next_action(self) -> ConfiguredAction:
        """next action to execute based on the configuration"""
        return self.configuration.next_action

    @property
    def status_code(self) -> int:
        """returns a status code representative of the state of the Action"""
        if self.error is True:
            code = -1
        elif self.time_started is None:
            code = 1
        elif self.time_completed is None:
            code = 2
        else:
            code = 3
        return code

    def as_dict(self) -> dict:
        """dictionary of relevant information"""
        out = {
            'name': self.configuration.name,
            'configuration': self.configuration.name,
            'arguments': self.args,
            'keyword_arguments': self.kwargs,
            'status_code': self.status_code,
            'status': self.status_string,
        }
        if self.time_started is not None:
            out['time_started'] = self.time_started
            out['timestamp'] = str(self.started_timestamp)
        if self.time_completed is not None:
            out['time_completed'] = self.time_completed
            out['duration'] = self.time_completed - self.time_started
            if self.error is True:
                out['error_during_execution'] = True
                out['error_details'] = str(self.error_details)
            else:  # if completed and no error, include the method return
                out['action_return'] = self.action_return
        return out

    def copy_configuration(self) -> "TrackedAction":
        """
        Copies the configuration of a TrackedAction instance.

        :return: TrackedAction instance with identical configuration
        """
        return self.__class__(
            self.configuration,
            *self.args,
            **self.kwargs,
        )

    @classmethod
    def create_from_configured(cls, source: Union[Action, ConfiguredAction]) -> "TrackedAction":
        """
        Creates a TrackedAction instance from a provided Action or ConfiguredAction

        :param source: action to generate from
        :return: instantiated TrackedAction
        """
        warnings.warn(
            'creating TrackedAction instances from ConfiguredAction instances is now the intended method of '
            'instantiation please update your code to instantiate directly',
            DeprecationWarning,
            stacklevel=2,
        )
        return cls(source)

    _ensure_unfrozen = staticmethod(_ensure_unfrozen)


class ThreadedTrackedAction(TrackedAction):
    UUID_PREFIX = 'tta-'

    def __init__(self,
                 configuration: Union[str, Callable, Action, ConfiguredAction],
                 *args,
                 **kwargs,
                 ):
        """
        A threaded version of a tracked action. This action may be executed asynchronously.

        :param configuration: action to perform (callable)
        :param Action, ConfiguredAction, next_step: for creating a sequence of actions, the next_action is the action
            that should be executed after this action has been executed
        :param args: arguments for the action
        :param kwargs: keyword arguments for the action
        """
        TrackedAction.__init__(
            self,
            configuration=configuration,
            *args,
            **kwargs,
        )
        self._lock = threading.Lock()
        self._thread: threading.Thread = None

    def _executor(self, *args, **kwargs):
        """target method for the internal thread"""
        with self._lock:
            super(ThreadedTrackedAction, self).__call__(*args, **kwargs)

    def _trigger(self, *args, **kwargs):
        """
        Triggers the action execution

        :param args: arguments for the call
        :param kwargs: kwargs for the call
        """
        if self.triggered is True:
            raise SystemError(f'multiple executions of the same {self.__class__.__name__} are not supported')
        logger.debug(f'triggering execution of {self.__str__()}')
        self.update_args(*args)
        self.update_kwargs(**kwargs)
        self._thread = threading.Thread(
            target=self._executor,
            name=f'{self.uuid} executor',
            daemon=True,
            args=self.args,
            kwargs=self.kwargs,
        )
        self._thread.start()

    def trigger(self, *args, **kwargs):
        """
        Triggers the action execution

        :param args: arguments for the call
        :param kwargs: kwargs for the call
        """
        self._trigger(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        self._trigger(*args, **kwargs)

    @property
    def in_progress(self) -> bool:
        """
        whether the action is in progress (the action is in the process of being executed)
        """
        if self.triggered is True:
            return self._lock.locked()
        return False

    @property
    def complete(self) -> bool:
        """whether the actions have completed"""
        if self.triggered is True:  # if instance has been triggered, check lock state
            return not self._lock.locked()
        return False

    def wait_for_completion(self):
        """
        Waits for the completion of the time point
        """
        if self.in_progress is True:
            logger.info('waiting for action to complete')
            self._thread.join()  # todo support timeout


class TrackedActionList(ActionList):
    def __init__(self, *actions: Union[Action, ConfiguredAction, TrackedAction]):
        """
        A manager for a list of Tracked Actions.

        :param actions: actions to sequence
        """
        self._action_list: List[TrackedAction] = []
        super().__init__(*actions)
        # todo define what happens when you copy the list (should create unexecuted trackedactions)

    @staticmethod
    def _ensure_type(action: Union[str, Action, ConfiguredAction, TrackedAction]) -> TrackedAction:
        """Ensures that the provided action is of the correct type for the list"""
        if isinstance(action, TrackedAction):
            return action
        else:
            return TrackedAction(action)



from .version import __version__
from .action import (Action, EvalAction, ActionList, ConfiguredAction, ConfiguredActionList, ConfiguredDecision,
                     TrackedDecision, TrackedAction)
from .sequencing import Automation, ExecutionSequence
from .timepoint import TimePoint, ActionTimePoint
from .scheduler import SamplingScheduler

__all__ = [
    'SamplingScheduler',
    'Automation',
    'ExecutionSequence',
    'TimePoint',
    'ActionTimePoint',
    'Action',
    'ActionList',
    'EvalAction',
    'ConfiguredAction',
    'ConfiguredActionList',
    'ConfiguredDecision',
    'TrackedAction',
    'TrackedDecision',
]



"""
Contains classes for managing and applying context to "actions" (functions).

`Action` is a function register which enables default arg/kwarg specification and retrieval of action by function name.

`ConfiguredAction` is an action that is configured to run (allows for multiple configurations of `Action` classes.

`TrackedAction` is an action that may be run with storage of context (duration, time started, function return,
args/kwargs that were used for execution, and error details if any).

"""

from .basic import Action, EvalAction, ActionList
from .configured import ConfiguredAction, ConfiguredActionList, ThreadedTrackedAction, TrackedActionList, TrackedAction
from .logic import ConfiguredDecision, TrackedDecision

__all__ = [
    'Action',
    'EvalAction',
    'ActionList',
    'ConfiguredAction',
    'ConfiguredActionList',
    'ConfiguredDecision',
    'TrackedAction',
    'TrackedDecision',
]

# todo figure out how to trigger a pre-trigger event (e.g. turn on instrument and condition)
#   - executed contextually
#   - or rely on the user to specify pre-events in provided methods



import logging

from typing import Union, Callable, Iterable

from .basic import Action
from .configured import ConfiguredAction, ConfiguredActionList, TrackedActionList, TrackedAction
from .operators import decision


logger = logging.getLogger(__name__)


class ConfiguredDecision(ConfiguredAction):
    UUID_PREFIX = 'cda-'

    def __init__(self,
                 *logic_checks: Union[Callable, Action, str, ConfiguredAction],
                 configuration_name: str = None,
                 **kwargs,
                 ):
        """
        Class for configuring a logic tree. The logic checks will iterate until an action return is True.

        :param logic_checks: logic check actions. These must return a bool
        :param configuration_name: name for the configuration
        """
        ConfiguredAction.__init__(
            self,
            action=decision,
            configuration_name=configuration_name,
            **kwargs,
        )
        self._logic_sequence: ConfiguredActionList = ConfiguredActionList()
        self.logic_sequence = logic_checks

    @property
    def logic_sequence(self) -> ConfiguredActionList:
        """list of logic check actions to be performed"""
        return self._logic_sequence

    @logic_sequence.setter
    def logic_sequence(self, value: Iterable):
        self._logic_sequence = ConfiguredActionList(*value)
        for action in self._logic_sequence:
            if action.action.returns_bool is False:
                logger.warning(
                    f'the action {action} does not type hint bool, using this action for decisions may result in '
                    f'unexpected errors'
                )

    def get_tracked_from_config(self, **kwargs) -> "TrackedDecision":
        """
        Generates a tracked instance from the configured action

        :param kwargs: additional kwargs that differ from the instance configuration
        :return: TrackedDecision instance.
        """
        return TrackedDecision(configuration=self, **kwargs)


class TrackedDecision(TrackedAction):
    UUID_PREFIX = 'da-'

    def __init__(self,
                 configuration: Union[str, ConfiguredDecision],
                 **kwargs,
                 ):
        """
        When called, given multiple ConfiguredActions, each of which their action must return a bool, sequentially run
        each ConfiguredAction, and for the first ConfiguredAction that returns True, return the next_action of the
        ConfiguredAction

        :param args: ConfiguredActions that have actions that return a bool
        :param configuration_name: optional name for the configuration (for key retrieval)
        """
        if isinstance(configuration, ConfiguredDecision) is False:
            configuration = ConfiguredDecision.class_instance_by_id(configuration)
        self._configuration: ConfiguredDecision = None

        TrackedAction.__init__(
            self,
            configuration=configuration,
            **kwargs,
        )
        self._next_action: ConfiguredAction = None

    def __call__(self, *args, **kwargs):
        logger.info('beginning execution of logic tree')
        # convert configured actions to tracked actions
        actions = TrackedActionList(*self.configuration.logic_sequence)
        super(TrackedDecision, self).__call__(
            *actions,
            **kwargs,
        )
        logger.info(f'logic execution complete')

    @property
    def configuration(self) -> ConfiguredDecision:
        """the configured sequence associated with the instance"""
        return self._configuration

    @configuration.setter
    def configuration(self, value: ConfiguredDecision):
        if isinstance(value, ConfiguredDecision) is False:
            raise TypeError(f'a ConfiguredDecision must be used as a configuration for {self.__class__.__name__}')
        self._configuration = value
        self.update_args(*self._configuration.args)
        self.update_kwargs(*self._configuration.kwargs)

    def copy_configuration(self) -> "TrackedDecision":  # todo update docstring
        """
        Copies the configuration of a ExecutionSequence instance.

        :return: ExecutionSequence instance with identical configuration
        """
        return self.__class__(
            self.configuration,
            **self.kwargs,
        )

    @property
    def next_action(self) -> ConfiguredAction:
        """next action as determined by the logic tree"""
        return self._action_return



import copy
import inspect
import logging
import warnings
from typing import List, Callable, Union, MutableSequence

from ..mixin import InstanceRegistry, ClassProperty


logger = logging.getLogger(__name__)


class Action(InstanceRegistry):
    UUID_PREFIX = 'action-'

    def __init__(self,
                 action: Callable,
                 *default_args,
                 **default_kwargs,
                 ):
        """
        Class for storing and tracking functions and methods to be executed.

        :param action: action method to execute when called
        :param default_args: default arguments for the action
        :param default_kwargs: default kwargs for the action
        """
        # todo consider implementing action removal
        if self.action_is_registered(action) is True:
            raise ValueError(
                f'The action "{action}" is already registered, use {self.__class__.__name__}.register_action instead'
            )
        InstanceRegistry.__init__(self)
        self.action = action
        self.__doc__ = inspect.getdoc(action)
        self.signature = inspect.signature(action)
        self.default_args = default_args
        self.logger = logger.getChild(f'{self.__class__.__name__}.{self.name}')
        fn_kwargs = {
            name: param.default
            for name, param in self.signature.parameters.items() if param.default is not inspect._empty
        }
        fn_kwargs.update(default_kwargs)
        self.default_kwargs = fn_kwargs

    def __repr__(self):
        return f'{self.__class__.__name__}({self.action.__name__})'

    def __call__(self, *args, **kwargs):
        self.logger.debug(f'calling {self.name} with args {args} and kwargs {kwargs}')
        return self.action(*args, **kwargs)

    @property
    def name(self) -> str:
        """action method name"""
        return self.action.__name__

    @property
    def return_type(self):
        """the return type of the action"""
        if self.signature.return_annotation is inspect._empty:
            return None
        return self.signature.return_annotation

    @classmethod
    def action_is_registered(cls,
                             action: Union[str, Callable, "Action"]
                             ) -> bool:
        """
        Checks whether the provided action is registered as a Action instance.

        :param action: action to check
        :return: whether action is registered
        """
        for instance in cls.class_registry:
            if any([
                instance.name == action,
                instance == action,
                instance.action == action,
            ]):
                return True
        return False

    @classmethod
    def register_action(cls,
                        action: Union[str, Callable, "Action"],
                        ) -> "Action":
        """
        Registers a new instance or returns an existing instance if the action is already registered.

        :param action: action method
        :return: Action instance
        """
        # todo consider logging that a new action has been registered
        # catch class instance
        if isinstance(action, Action):
            return action
        elif hasattr(action, 'action') and isinstance(action.action, Action):
            return action.action
        # search for action name
        elif type(action) is str:
            for instance in cls.class_registry:
                if instance.name == action:
                    return instance
            # if no action matched, raise
            raise ValueError(f'The action "{action}" is not registered')
        else:
            # check for preexisting actions
            for instance in cls.class_registry:
                if instance.name == action.__name__:
                    return instance
            # if not found, create and return
            return cls(
                action
            )

    @classmethod
    def register_actions(cls, *actions: Union[str, Callable, "Action"]) -> List["Action"]:
        """
        Registers multiple actions at once

        :param actions: action methods to be registered
        :return: list of actions which were registered
        """
        return [
            cls.register_action(action) for action in actions
        ]

    @classmethod
    def get_registered_actions(cls) -> List["Action"]:
        """returns a list of registered Action instances"""
        warnings.warn(
            'get_registered_actions has been deprecated, retrieve registered_actions directly',
            DeprecationWarning,
            stacklevel=2,
        )
        return cls.registered_actions

    @ClassProperty
    def registered_actions(cls) -> List["Action"]:
        """a list of registered Action instances"""
        return cls.class_registry

    @classmethod
    def get_registered_action_names(cls) -> List[str]:
        """returns a list of the names of the registered Scheduler Actions"""
        warnings.warn(
            'get_registered_action_names has been deprecated, retrieve registered_action_names directly',
            DeprecationWarning,
            stacklevel=2,
        )
        return cls.registered_action_names

    @ClassProperty
    def registered_action_names(cls) -> List[str]:
        """a list of the names of the registered Actions"""
        return [action.name for action in cls.registered_actions]

    @property
    def returns_bool(self) -> bool:
        """whether the action return type bool"""
        return self.return_type is bool


class EvalAction(Action):
    def __init__(self,
                 action: Callable,
                 *default_args,
                 **default_kwargs,
                 ):
        """
        Subclass of an Action where the action must return a bool. Note that this class does not enforce this, but
        warns the user if a non-bool type hint was detected.

        :param action: action method to execute when called
        :param default_args: default arguments for the action
        :param default_kwargs: default kwargs for the action
        """
        super().__init__(
            action=action,
            *default_args,
            **default_kwargs,
         )
        if self.return_type is not bool:
            warnings.warn(f'The action "{self.name}" provided as an EvalAction does not have a bool return type')


class ActionList(MutableSequence):
    def __init__(self, *actions: Action):
        """
        A manager for a list of Actions.

        :param actions: actions to sequence
        """
        self._action_list: List[Action] = []
        for action in actions:
            self.append(action)

    def __repr__(self):
        return f'{self.__class__.__name__}[{", ".join([str(action) for action in self._action_list])}]'

    @staticmethod
    def _ensure_type(action: Union[str, Action]) -> Action:
        """Ensures that the provided action is of the correct type for the list"""
        return Action.register_action(action)

    def __getitem__(self, item) -> Action:
        return self._action_list[item]

    def __setitem__(self, key, value: Action):
        value = self._ensure_type(value)
        self._action_list[key] = value

    def __delitem__(self, key):
        del self._action_list[key]

    def __len__(self):
        return len(self._action_list)

    def __copy__(self):
        # provides a copy of the class while preventing mutation of the actions in the instance itself
        return self.__class__(
            *[copy.copy(action) for action in self]
        )

    def insert(self, index: int, action: Action) -> None:
        """
        Inserts the provided action into the Action list

        :param index: index to insert at
        :param action: action object to insert
        :return:
        """
        action = self._ensure_type(action)
        self._action_list.insert(
            index,
            action,
        )